@@ -52,7 +52,7 @@
 */
 
 #ifndef gmic_version
-#define gmic_version 296
+#define gmic_version 315
 
 #ifndef gmic_pixel_type
 #define gmic_pixel_type float
@@ -74,7 +74,11 @@
 #endif // #if cimg_OS==2
 #endif // #ifndef gmic_uint64
 
-#ifndef gmic_build
+// Define some special character codes used for replacement in double quoted strings.
+const char gmic_dollar = 23, gmic_lbrace = 24, gmic_rbrace = 25, gmic_comma = 26, gmic_dquote = 28,
+  gmic_store = 29; // <- this one is only used in variable names.
+
+#ifndef gmic_core
 
 // Define classes 'gmic_image<T>' and 'gmic_list<T>'.
 //---------------------------------------------------
@@ -173,7 +177,7 @@ namespace cimg_library {
 #undef gmic_list
 #endif // #ifndef cimg_version
 
-#else // #ifndef gmic_build
+#else // #ifndef gmic_core
 
 // Define private functions, used to compile libgmic.
 //---------------------------------------------------
@@ -191,40 +195,53 @@ namespace cimg_library {
 #ifdef cimg_version
 #error "[gmic] *** Error *** File 'CImg.h' has been already included (should have been done first in file 'gmic.h')."
 #endif
+#ifndef cimg_plugin
 #define cimg_plugin "gmic.cpp"
+#endif
+#ifndef cimglist_plugin
 #define cimglist_plugin "gmic.cpp"
-
-#ifdef cimg_use_abort
-inline bool *gmic_abort_ptr(bool *const p_is_abort);
-#define cimg_abort_init bool *const gmic_is_abort = ::gmic_abort_ptr(0)
-#define cimg_abort_test if (*gmic_is_abort) throw CImgAbortException()
 #endif
 
+#ifdef cimg_use_abort
+inline bool *gmic_current_is_abort();
+#define cimg_abort_init \
+  bool *const gmic_is_abort = ::gmic_current_is_abort()
+#define cimg_abort_test \
+  if (*gmic_is_abort) throw CImgAbortException()
+#endif
+
+inline double gmic_mp_dollar(const char *const str, void *const p_list);
+#define cimg_mp_operator_dollar(str) \
+  ::gmic_mp_dollar(str,&imglist)
+
 template<typename T>
-inline double gmic_mp_run(char *const str,
+inline double gmic_mp_get(double *const ptrd, const unsigned int siz, const bool to_string, const char *const str,
                           void *const p_list, const T& pixel_type);
+#define cimg_mp_func_get(ptrd,siz,to_string,str) \
+  return ::gmic_mp_get(ptrd,siz,to_string,str,&mp.imglist,(T)0)
+
+inline double gmic_mp_set(const double *const ptrs, const unsigned int siz, const char *const str,
+                          void *const p_list);
+#define cimg_mp_func_set(ptrs,siz,str) \
+  return ::gmic_mp_set(ptrs,siz,str,&mp.imglist)
+
+inline double gmic_mp_name(const unsigned int ind, double *const out_str, const unsigned int siz,
+                           void *const p_list);
+#define cimg_mp_func_name(ind,out_str,siz) \
+  return ::gmic_mp_name(ind,out_str,siz,&mp.imglist)
+
+template<typename T>
+inline double gmic_mp_run(char *const str, void *const p_list, const T& pixel_type);
 #define cimg_mp_func_run(str) \
-  return ::gmic_mp_run(str,&mp.listout,(T)0)
+  return ::gmic_mp_run(str,&mp.imglist,(T)0)
 
-template<typename Ts, typename T>
-inline double gmic_mp_get(Ts *const ptr, const unsigned int siz, const bool to_numbers, const char *const str,
-                          void *const p_list, const T& pixel_type);
-#define cimg_mp_func_get(ptr,siz,to_numbers,str) \
-  return ::gmic_mp_get(ptr,siz,to_numbers,str,&mp.listout,(T)0)
-
-template<typename Ts, typename T>
-inline double gmic_mp_store(const Ts *const ptr,
+template<typename T>
+inline double gmic_mp_store(const double *const ptrs, const unsigned int siz,
                             const unsigned int w, const unsigned int h, const unsigned int d, const unsigned int s,
                             const bool is_compressed, const char *const str,
                             void *const p_list, const T& pixel_type);
-#define cimg_mp_func_store(ptr,w,h,d,s,is_compressed,str) \
-  return ::gmic_mp_store(ptr,w,h,d,s,is_compressed,str,&mp.listout,(T)0)
-
-template<typename T, typename Ts>
-inline double gmic_mp_name(const unsigned int ind, Ts *const out_str, const unsigned int siz,
-                           void *const p_list, const T& pixel_type);
-#define cimg_mp_func_name(ind,out_str,siz) \
-  return ::gmic_mp_name(ind,out_str,siz,&mp.listout,(T)0)
+#define cimg_mp_func_store(ptrs,siz,w,h,d,s,is_compressed,str) \
+  return ::gmic_mp_store(ptrs,siz,w,h,d,s,is_compressed,str,&mp.imglist,(T)0)
 
 #ifndef cimg_display
 #define cimg_display 0
@@ -245,12 +262,7 @@ inline double gmic_mp_name(const unsigned int ind, Ts *const out_str, const unsi
 #include <signal.h>
 
 #endif // #if cimg_OS==2
-
-// Define some special character codes used for replacement in double quoted strings.
-const char gmic_dollar = 23, gmic_lbrace = 24, gmic_rbrace = 25, gmic_comma = 26, gmic_dquote = 28,
-  gmic_store = 29; // <- this one is only used in variable names.
-
-#endif // #ifndef gmic_build
+#endif // #ifndef gmic_core
 
 // Define main libgmic class 'gmic'.
 //----------------------------------
@@ -267,26 +279,32 @@ struct gmic {
   gmic();
 
   template<typename T>
-  gmic(const char *const commands_line,
-       const char *const custom_commands=0,
-       const bool include_stdlib=true,
-       float *const p_progress=0, bool *const p_is_abort=0,
-       const T& pixel_type=(T)0);
+  gmic(const char *const commands_line, const char *const custom_commands=0, const bool include_stdlib=true,
+       float *const p_progress=0, bool *const p_is_abort=0, const T& pixel_type=(T)0);
 
   template<typename T>
   gmic(const char *const commands_line,
        gmic_list<T>& images, gmic_list<char>& images_names, const char *const custom_commands=0,
        const bool include_stdlib=true, float *const p_progress=0, bool *const p_is_abort=0);
 
+  gmic& assign();
+
+  template<typename T>
+  gmic& assign(const char *const commands_line, const char *const custom_commands=0, const bool include_stdlib=true,
+               float *const p_progress=0, bool *const p_is_abort=0, const T& pixel_type=(T)0);
+
+  template<typename T>
+  gmic& assign(const char *const commands_line,
+               gmic_list<T>& images, gmic_list<char>& images_names, const char *const custom_commands=0,
+               const bool include_stdlib=true, float *const p_progress=0, bool *const p_is_abort=0);
+
   // Run G'MIC pipeline on an already-constructed object.
   template<typename T>
-  gmic& run(const char *const commands_line,
-            float *const p_progress=0, bool *const p_is_abort=0,
+  gmic& run(const char *const commands_line, float *const p_progress=0, bool *const p_is_abort=0,
             const T& pixel_type=(T)0);
 
   template<typename T>
-  gmic& run(const char *const commands_line,
-            gmic_list<T> &images, gmic_list<char> &images_names,
+  gmic& run(const char *const commands_line, gmic_list<T> &images, gmic_list<char> &images_names,
             float *const p_progress=0, bool *const p_is_abort=0);
 
   // These functions return (or init) G'MIC-specific paths.
@@ -297,23 +315,22 @@ struct gmic {
   // Functions below should be considered as *private*, and should not be used in user's code.
   template<typename T>
   static bool search_sorted(const char *const str, const T& list, const unsigned int length, unsigned int &out_ind);
+  static const gmic_image<void*> current_run(const char *const func_name, void *const p_list);
+  static double mp_dollar(const char *const str, void *const p_list);
   template<typename T>
-  static double mp_run(char *const str,
+  static double mp_get(double *const ptrd, const unsigned int siz, const bool to_string, const char *const str,
                        void *const p_list, const T& pixel_type);
-  template<typename Ts, typename T>
-  static double mp_get(Ts *const ptr, const unsigned int siz, const bool to_numbers, const char *const str,
-                       void *const p_list, const T& pixel_type);
-  template<typename Ts, typename T>
-  static double mp_store(const Ts *const ptr,
+  static double mp_set(const double *const ptrs, const unsigned int siz, const char *const str, void *const p_list);
+  static double mp_name(const unsigned int ind, double *const out_str, const unsigned int siz, void *const p_list);
+  template<typename T>
+  static double mp_run(char *const str, void *const p_list, const T& pixel_type);
+  template<typename T>
+  static double mp_store(const double *const ptrs, const unsigned int siz,
                          const unsigned int w, const unsigned int h, const unsigned int d, const unsigned int s,
                          const bool is_compressed, const char *const str,
                          void *const p_list, const T& pixel_type);
-  template<typename T, typename Ts>
-  static double mp_name(const unsigned int ind, Ts *const out_str, const unsigned int siz,
-                        void *const p_list, const T& pixel_type);
   static bool get_debug_info(const char *const s, unsigned int &line_number, unsigned int &file_number);
-  static int _levenshtein(const char *const s, const char *const t,
-                          gmic_image<int>& d, const int i, const int j);
+  static int _levenshtein(const char *const s, const char *const t, gmic_image<int>& d, const int i, const int j);
   static int levenshtein(const char *const s, const char *const t);
   static unsigned int hashcode(const char *const str, const bool is_variable);
   static bool command_has_arguments(const char *const command);
@@ -322,29 +339,24 @@ struct gmic {
   static char *strreplace_bw(char *const str);
   static unsigned int strescape(const char *const str, char *const res);
   static const gmic_image<char>& decompress_stdlib();
-  static bool *abort_ptr(bool *const p_is_abort);
 
   template<typename T>
-  void _gmic(const char *const commands_line,
-             gmic_list<T>& images, gmic_list<char>& images_names,
-             const char *const custom_commands, const bool include_stdlib,
-             float *const p_progress, bool *const p_is_abort);
+  gmic& _gmic(const char *const commands_line, gmic_list<T>& images, gmic_list<char>& images_names,
+              const char *const custom_commands, const bool include_stdlib,
+              float *const p_progress, bool *const p_is_abort);
 
-  gmic_image<char> get_variable(const char *const name,
-                                const unsigned int *const variables_sizes=0,
+  gmic_image<char> get_variable(const char *const name, const unsigned int *const variables_sizes=0,
                                 const gmic_list<char> *const images_names=0) const;
-  const char *set_variable(const char *const name, const char *const value,
-                           const char operation='=',
-                           const unsigned int *const variables_sizes=0);
+  const char *set_variable(const char *const name, const char operation='=', const char *const value=0,
+                           const double *const pvalue=0, const unsigned int *const variables_sizes=0);
   const char *set_variable(const char *const name, const gmic_image<unsigned char>& value,
                            const unsigned int *const variables_sizes=0);
 
   gmic& add_commands(const char *const data_commands, const char *const commands_file=0,
-                     unsigned int *count_new=0, unsigned int *count_replaced=0,
-                     bool *const is_entrypoint=0);
-  gmic& add_commands(std::FILE *const file, const char *const filename=0,
-                     unsigned int *count_new=0, unsigned int *count_replaced=0,
-                     bool *const is_entrypoint=0);
+                     const bool add_debug_info=false, unsigned int *count_new=0,
+                     unsigned int *count_replaced=0, bool *const is_entrypoint=0);
+  gmic& add_commands(std::FILE *const file, const char *const filename=0, const bool add_debug_info=false,
+                     unsigned int *count_new=0, unsigned int *count_replaced=0, bool *const is_entrypoint=0);
 
   gmic_image<char> callstack2string(const bool _is_debug=false) const;
   gmic_image<char> callstack2string(const gmic_image<unsigned int>& callstack_selection,
@@ -357,29 +369,24 @@ struct gmic {
                                           const gmic_list<char>& names, const char *const command,
                                           const bool is_selection=true);
 
-  gmic_image<char>& selection2string(const gmic_image<unsigned int>& selection,
-                                     const gmic_list<char>& images_names,
-                                     const unsigned int display_selection,
-                                     gmic_image<char>& res) const;
+  gmic_image<char>& selection2string(const gmic_image<unsigned int>& selection, const gmic_list<char>& images_names,
+                                     const unsigned int display_selection, gmic_image<char>& res) const;
 
   gmic_list<char> commands_line_to_CImgList(const char *const commands_line);
 
   template<typename T>
-  void _gmic_substitute_args(const char *const argument, const char *const argument0,
-                             const char *const command, const char *const item,
-                             const gmic_list<T>& images);
+  void _gmic_substitute_args(const char *const argument, const char *const argument0, const char *const command,
+                             const char *const item, const gmic_list<T>& images);
 
   gmic& print(const char *format, ...);
   gmic& error(const bool output_header, const char *format, ...);
   gmic& debug(const char *format, ...);
 
   template<typename T>
-  gmic_image<char> substitute_item(const char *const source,
-                                   gmic_list<T>& images, gmic_list<char>& images_names,
+  gmic_image<char> substitute_item(const char *const source, gmic_list<T>& images, gmic_list<char>& images_names,
                                    gmic_list<T>& parent_images, gmic_list<char>& parent_images_names,
                                    const unsigned int *const variables_sizes,
-                                   const gmic_image<unsigned int> *const command_selection,
-                                   const bool is_image_expr);
+                                   const gmic_image<unsigned int> *const command_selection, const bool is_image_expr);
 
   template<typename T>
   void wait_threads(void *const p_gmic_threads, const bool try_abort, const T& pixel_type);
@@ -404,29 +411,20 @@ struct gmic {
   gmic& debug(const gmic_list<T>& list, const char *format, ...);
 
   template<typename T>
-  gmic& print_images(const gmic_list<T>& images,
-                     const gmic_list<char>& images_names,
-                     const gmic_image<unsigned int>& selection,
-                     const bool is_header=true);
+  gmic& print_images(const gmic_list<T>& images, const gmic_list<char>& images_names,
+                     const gmic_image<unsigned int>& selection, const bool is_header=true);
   template<typename T>
-  gmic& display_images(const gmic_list<T>& images,
-                       const gmic_list<char>& images_names,
-                       const gmic_image<unsigned int>& selection,
-                       unsigned int *const XYZ,
+  gmic& display_images(const gmic_list<T>& images, const gmic_list<char>& images_names,
+                       const gmic_image<unsigned int>& selection, unsigned int *const XYZ,
                        const bool exit_on_anykey);
   template<typename T>
-  gmic& display_plots(const gmic_list<T>& images,
-                      const gmic_list<char>& images_names,
-                      const gmic_image<unsigned int>& selection,
-                      const unsigned int plot_type, const unsigned int vertex_type,
-                      const double xmin, const double xmax,
-                      const double ymin, const double ymax,
-                      const bool exit_on_anykey);
+  gmic& display_plots(const gmic_list<T>& images, const gmic_list<char>& images_names,
+                      const gmic_image<unsigned int>& selection, const unsigned int plot_type,
+                      const unsigned int vertex_type, const double xmin, const double xmax,
+                      const double ymin, const double ymax, const bool exit_on_anykey);
   template<typename T>
-  gmic& display_objects3d(const gmic_list<T>& images,
-                          const gmic_list<char>& images_names,
-                          const gmic_image<unsigned int>& selection,
-                          const gmic_image<unsigned char>& background3d,
+  gmic& display_objects3d(const gmic_list<T>& images, const gmic_list<char>& images_names,
+                          const gmic_image<unsigned int>& selection, const gmic_image<unsigned char>& background3d,
                           const bool exit_on_anykey);
   template<typename T>
   gmic_image<T>& check_image(const gmic_list<T>& list, gmic_image<T>& img);
@@ -434,43 +432,37 @@ struct gmic {
   const gmic_image<T>& check_image(const gmic_list<T>& list, const gmic_image<T>& img);
 
   template<typename T>
-  gmic& remove_images(gmic_list<T>& images, gmic_list<char>& images_names,
-                      const gmic_image<unsigned int>& selection,
+  gmic& remove_images(gmic_list<T>& images, gmic_list<char>& images_names, const gmic_image<unsigned int>& selection,
                       const unsigned int start, const unsigned int end);
 
   template<typename T>
-  gmic& _run(const gmic_list<char>& commands_line,
-             gmic_list<T> &images, gmic_list<char> &images_names,
+  gmic& _run(const gmic_list<char>& commands_line, gmic_list<T> &images, gmic_list<char> &images_names,
              float *const p_progress, bool *const p_is_abort);
 
   template<typename T>
-  gmic& _run(const gmic_list<char>& commands_line, unsigned int& position,
-             gmic_list<T>& images, gmic_list<char>&images_names,
-             gmic_list<T>& parent_images, gmic_list<char>& parent_images_names,
-             const unsigned int *const variables_sizes,
-             bool *const is_noargs, const char *const parent_arguments,
+  gmic& _run(const gmic_list<char>& commands_line, unsigned int& position, gmic_list<T>& images,
+             gmic_list<char>&images_names, gmic_list<T>& parent_images, gmic_list<char>& parent_images_names,
+             const unsigned int *const variables_sizes, bool *const is_noargs, const char *const parent_arguments,
              const gmic_image<unsigned int> *const command_selection);
 
-  // Class variables.
+  // Class attributes.
   static const char *builtin_commands_names[];
   static gmic_image<int> builtin_commands_inds;
   static gmic_image<char> stdlib;
-  static gmic_list<void*> list_p_is_abort;
   static bool is_display_available;
 
-  gmic_list<char> *commands, *commands_names, *commands_has_arguments,
-    *_variables, *_variables_names, **variables, **variables_names,
-    commands_files, callstack;
-  gmic_image<unsigned int> dowhiles, fordones, repeatdones;
+  gmic_list<char> *commands, *commands_names, *commands_has_arguments, *_variables, *_variables_names,
+    **variables, **variables_names, commands_files, callstack;
+  gmic_image<unsigned int> dowhiles, fordones, foreachdones, repeatdones;
   gmic_image<unsigned char> light3d;
   gmic_image<void*> display_windows;
   gmic_image<char> status;
 
   float focale3d, light3d_x, light3d_y, light3d_z, specular_lightness3d, specular_shininess3d, _progress, *progress;
   gmic_uint64 reference_time;
-  unsigned int nb_dowhiles, nb_fordones, nb_repeatdones, nb_carriages_default, nb_carriages_stdout,
+  unsigned int nb_dowhiles, nb_fordones, nb_foreachdones, nb_repeatdones, nb_carriages_default, nb_carriages_stdout,
     debug_filename, debug_line, cimg_exception_mode;
-  int verbosity,render3d, renderd3d, network_timeout;
+  int verbosity, render3d, renderd3d, network_timeout;
   bool allow_entrypoint, is_change, is_debug, is_running, is_start, is_return, is_quit, is_double3d, is_debug_info,
     _is_abort, *is_abort, is_abort_thread;
   const char *starting_commands_line;
@@ -505,34 +497,46 @@ struct gmic_exception {
   }
 };
 
+inline bool *gmic_current_is_abort() {
+  static bool def = false;
+  gmic_image<void*> gr = gmic::current_run("gmic_abort_init()",0);
+  return gr?((gmic*)(gr[0]))->is_abort:&def;
+}
+
+inline double gmic_mp_dollar(const char *const str, void *const p_list) {
+  return gmic::mp_dollar(str,p_list);
+}
+
+template<typename T>
+inline double gmic_mp_get(double *const ptrd, const unsigned int siz, const bool to_string, const char *const str,
+                          void *const p_list, const T& pixel_type) {
+  return gmic::mp_get(ptrd,siz,to_string,str,p_list,pixel_type);
+}
+
+inline double gmic_mp_set(const double *const ptrs, const unsigned int siz, const char *const str,
+                          void *const p_list) {
+  return gmic::mp_set(ptrs,siz,str,p_list);
+}
+
+inline double gmic_mp_name(const unsigned int ind, double *const out_str, const unsigned int siz,
+                           void *const p_list) {
+  return gmic::mp_name(ind,out_str,siz,p_list);
+}
+
 template<typename T>
 inline double gmic_mp_run(char *const str,
                           void *const p_list, const T& pixel_type) {
   return gmic::mp_run(str,p_list,pixel_type);
 }
 
-template<typename Ts, typename T>
-inline double gmic_mp_get(Ts *const ptr, const unsigned int siz, const bool to_numbers, const char *const str,
-                          void *const p_list, const T& pixel_type) {
-  return gmic::mp_get(ptr,siz,to_numbers,str,p_list,pixel_type);
-}
-
-template<typename Ts, typename T>
-inline double gmic_mp_store(const Ts *const ptr,
+template<typename T>
+inline double gmic_mp_store(const double *const ptrs, const unsigned int siz,
                             const unsigned int w, const unsigned int h, const unsigned int d, const unsigned int s,
                             const bool is_compressed, const char *const str,
                             void *const p_list, const T& pixel_type) {
-  return gmic::mp_store(ptr,w,h,d,s,is_compressed,str,p_list,pixel_type);
+  return gmic::mp_store(ptrs,siz,w,h,d,s,is_compressed,str,p_list,pixel_type);
 }
 
-template<typename T, typename Ts>
-inline double gmic_mp_name(const unsigned int ind, Ts *const out_str, const unsigned int siz,
-                           void *const p_list, const T& pixel_type) {
-  return gmic::mp_name(ind,out_str,siz,p_list,pixel_type);
-}
-
-inline bool *gmic_abort_ptr(bool *const p_is_abort) { return gmic::abort_ptr(p_is_abort); }
-
 #endif // #ifndef gmic_version
 
 // Local Variables:
